package fixedtime;
use 5.010;         # this is a user-defined pragma and needs perl 5.10 or higher
use warnings;
use strict;

our $VERSION = 0.01;

sub import   {
    shift;
    my %args = @_;
    $^H{fixedtime} = exists $args{stamp}
        ? $args{stamp} // CORE::time
        : CORE::time;
}

sub unimport { $^H{fixedtime} = undef }

sub epoch_offset {
    my $level = shift // 0;
    my $ctrl_h = ( caller $level + 1 )[10];
    return $ctrl_h->{fixedtime};
}

# redefine the time related functions
# this works because:
#   * pragma in effect     -> fixedtime::epoch_offset() is defined
#   * pragma not in effect -> fixedtime::epoch_offset() is not defined
#   * the // makes sure that for undef CORE::time is used
# NB: for gmtime and localtime:
#       when an epoch offset is passed, normal operation is in effect
BEGIN {
    *CORE::GLOBAL::time = sub {
        return fixedtime::epoch_offset() // CORE::time;
    };

    *CORE::GLOBAL::gmtime = sub (;$) {
        my $stamp = shift // fixedtime::epoch_offset() // CORE::time;
        CORE::gmtime( $stamp );
    };

    *CORE::GLOBAL::localtime = sub (;$) {
        my $stamp = shift // fixedtime::epoch_offset() // CORE::time;
        CORE::localtime( $stamp );
    };
}

1;

__END__

=head1 NAME

fixedtime - lexical pragma to fix the epoch offset for time related functions

=head1 SYNOPSIS

    use Test::More 'no_plan';

    my $nowstamp = time;
    my $fixstamp;
    {
        use fixedtime stamp => 1204286400; # 29 Feb 2008 12:00:00 GMT

        $fixstamp = time;
        is $fixstamp, 1204286400, "Fixed point in time ($fixstamp)";
        is scalar gmtime, "Fri Feb 29 12:00:00 2008", "@{[ scalar gmtime ]}";

        no fixedtime;
        is time, $nowstamp, "we ran fast enough ($nowstamp)";
    }

    is time, $nowstamp, "we ran fast enough ($nowstamp)";

=head1 DESCRIPTION

This pragma demonstrates the new perl 5.10 user-defined lexical pragma
capability. It uses the C<$^H{fixedtime}> hintshash entry to store the
epochoffset. Whenever C<$^H{fixedtime}> is undefined, the praga is
assumed not to be in effect.

The C<fixedtime> pragma affects L<time()>, L<gmtime()> and
L<localtime()> only when called without an argument.

=head2 use fixedtime [stamp => epoch_offset];

This will enable the pragma in the current lexical scope. When the
stamp argument is omitted, C<CORE::time()> is taken. While the pragma
is in effect the epochoffset is not changed.

B<Warning>: If you use a variable to set the epoch offset, make sure
it is initialized at compile time.

    my $stamp = 1204286400;
    use fixedtime stamp => $stamp; # Will not work as expected

You will need something like:

    use constant STAMP => 1204286400;
    use fixedtime stamp => STAMP;

=head2 no fixedtime;

This will disable the pragma in the current lexical scope.

=head1 SEE ALSO

L<perlpragma>

=head1 AUTHOR AND COPYRIGHT

(c) MMVIII All rights reserved, Abe Timmerman <abeltje@cpan.org>

=cut
