package Hugeint;
use 5.006;
use warnings;
use strict;

our $VERSION = 0.01;

use base 'Exporter';

use Hugeint::ByteArray;
use warnings::register;
use overload
    q{==}     => sub { hi_compare( @_ ) == 0 },
    q{""}     => sub { $_[0]->as_string      },
    q{=}      => sub { $_[0]->copy           },

# Bit operators
    q{neg}    => sub { $_[0]->copy->negate     },
    q{~}      => sub { $_[0]->copy->bit_negate },
    q{!}      => sub { $_[0]->is_zero        },

# These do their own argument swapping and use copies
    q{<=>}    => \&hi_compare,
    q{+}      => \&hi_add,
    q{-}      => \&hi_subtract,
    q{*}      => \&hi_multiply,
    q{/}      => \&hi_divide,
    q{%}      => \&hi_modulo,
    q{**}     => \&hi_power,
    q{&}      => \&hi_bit_and,
    q{|}      => \&hi_bit_or,
    q{^}      => \&hi_bit_xor,

# For optimizing
    q{+=}     => sub { $_[0]->add( $_[1] )           },
    q{-=}     => sub { $_[0]->subtract( $_[1] )      },
    q{*=}     => sub { $_[0]->multiply( $_[1] )      },
    q{/=}     => sub { $_[0]->divide( $_[1] ); $_[0] },
    q{%=}     => sub { $_[0]->divide( $_[1] )        },
    q{**=}    => sub { $_[0]->power( $_[1] )         },
    q{&=}     => sub { $_[0]->bit_and( $_[1] )       },
    q{|=}     => sub { $_[0]->bit_or( $_[1] )        },
    q{^=}     => sub { $_[0]->bit_xor( $_[1] )       },

    fallback => 1;

use constant MINPOW  => 1;
use constant MINSIZE => 2 ** MINPOW;

sub import {
    shift;
    @_ or return;
    @_ == 1 && $_[0] eq ':constant' or die "bad import '@_'";

    # warn "Start overloading numerical constants...";
    overload::constant(
        integer => sub { Hugeint->new( shift ) }
    );
}

sub unimport {
    # warn "Stop overloading numerical constants...";
    overload::remove_constant( 'integer' );
}

=head1 NAME

Hugeint - Arbitrary sized signed integers (2-complement little-endian)

=head1 SYNOPSIS

    use Hugeint;

    my $hi = Hugeint->new( '1234567890' );
    $hi->add( '987654321' );

    printf "Result is %s\n", $hi->as_string;

    print "Result is $hi\n"; # use overloading

=head1 DESCRIPTION

=head2 Hugeint->new( [$str_value] )

Create a C<Hugeint> object and initialize the instance with
C<$str_value>. If the initial value is omitted, it will be initialized
with zero (0).

=cut

sub new {
    my( $class, $value ) = @_;
    defined $value or $value = '0';

    tie my @store, 'Hugeint::ByteArray';
    my $self = bless \@store, $class;
    $self->assign( $value );
}

=head2 $hi->copy

Returns a copy of the instance.

=cut

sub copy {
    my( $self ) = @_;
    my $class = ref $self;
    my $copy = $class->new;
    @$copy = @$self;
    return $copy;
}

=head1 BASIC REGISTER OPERATIONS

=head2 $hi->negate

Changes the value of the instance with the effect of a unary minus:
    
    $hi->negate; # $hi = - $hi

=cut

sub negate {
    my( $self ) = @_;

    my $reg = 1;
    for my $byte ( @$self ) {
        $reg += ord( ~ chr( $byte ) );
        $byte = $reg % 256;
        $reg  = int $reg / 256;
    }
    $self->_pack;
}

=head2 $hi->bit_negate

c<bit_negate()> does a bitwise negation on it self:

    $hi->bit_negate; # $hi = ~ $hi

=cut

sub bit_negate {
    my( $self ) = @_;

    for my $byte ( @$self ) {
        $byte = ord( ~ chr( $byte ) );
    }

    return $self;
}

=head2 $hi->shift_left( $digits[, $keepsize] )

C<shift_left()> shifts the B<bytes> in the bytearray C<$digits>
positions "to the left".

    $hi->shift_left( 1 ); # $hi *= 256;

The C<$keepsize> argument makes sure the size of the bytearray is not
changed. This can be tricky as it can change the sign.

=cut

sub shift_left {
    my( $self, $digits, $keepsize ) = @_;
    unshift @$self, 0 for 1..$digits;
    $keepsize and pop @$self for 1..$digits;
    $self;
}

=head2 $hi->shift_right( $digits[, $keepsize] )

C<shift_right()> shifts the B<bytes> in the bytearray C<$digits>
positions "to the right".

    $hi->shift_right( 1 ); # $hi /= 256;

The C<$keepsize> argument makes sure the size of the bytearray is not
changed.

=cut

sub shift_right {
    my( $self, $digits, $keepsize ) = @_;
    $keepsize and push @$self, 0 for 1..$digits;
    @$self == 1 and push @$self, 0;
    shift @$self for 1..$digits;
    $self;
}

=begin private

=head2 $hi->_grow( $to_size )

This method extends the ByteArray that holds the internal representation.

=end private

=cut

sub _grow {
    my( $self, $size ) = @_;
    my $grow = $size - scalar @$self;
    $grow <= 0 and return $self;

    my @insert = ( $self->is_negative ? 255 : 0 ) x $grow;
    push @$self, @insert;

    $self;
}

=begin private

=head2 $hi->_pack( )
    
This method tries to optimize the internal size of each
Hugeint. Growing and shrinking is done to powers of 2 (so a 32byte
Hugeint will be upgraded to 64byte).

=end private

=cut

sub _pack {
    my $self = shift;
    @$self or push @$self, ( 0 ) x MINSIZE;
    if ( $self->is_negative ) { # $self->[-1] is always 255!
        pop @$self while @$self > MINSIZE and $self->[-2] == 255;
    } else {
        pop @$self while @$self > MINSIZE and $self->[-1] == 0
                     and $self->[-2] < 128;
    }

    my $pow = int( log( scalar @$self ) / log( 2 ) + 0.99 ) || MINPOW;
    $pow < MINPOW and $pow = MINPOW; # rather defensive for MINPOW = 1

    return $self->_grow( 2**$pow );
}

=begin private

=head2 __equalize_size( $a1, $a2[, $extra_size] )

This method makes sure that both Hugeints have the same length in their
internal representation,
    
=end private
        
=cut                 
    
sub __equalize_size($$;$) {
    my $extra = $_[2] || 0;
    if ( @{ $_[0] } > @{ $_[1] } ) {
        $_[0]->_grow( $extra + @{ $_[0] } );
        $_[1]->_grow( $extra + @{ $_[0] } );
    } elsif ( @{ $_[0] } < @{ $_[1] } ) {
        $_[0]->_grow( $extra + @{ $_[1] } );
        $_[1]->_grow( $extra + @{ $_[1] } );
    } elsif ( $extra > 0 ) {
        $_[0]->_grow( $extra + @{ $_[0] } );
        $_[1]->_grow( $extra + @{ $_[1] } );
    }
}

=head2 $hi->Dump

C<Dump()> Dumps some information about this Hugeint instance.

=cut

sub Dump {
    my( $self ) = @_;

    my $internal = tied( @$self );
    my $decimal  = $self->as_string;
    print <<EOI;
Internal size: @{[ length $$internal ]}
 Decimal size: @{[ length $decimal ]}
    as_string: $decimal
    as_binary: @{[ $self->as_binary ]}
EOI
}

=head1 ARITHMATIC OPERATIONS

=head2 $hi->bit_and( $value )

C<bit_and()> does a bitwise and on itself and C<$value>.

    $hi->bit_and( 256 ); # $hi &= 256

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub bit_and {
    my( $self, $value ) = @_;

    my $toand = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );
    __equalize_size( $self, $toand );

    for (my $i = 0; $i < @$self; $i++ ) {
        $$self[ $i ] &= $$toand[ $i ];
    }

    return $self->_pack;
}

=head2 $hi->bit_or( $value )

C<bit_or()> does a bitwise or on itself and C<$value>.

    $hi->bit_or( 127 ); # $hi |= 127

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub bit_or {
    my( $self, $value ) = @_;

    my $toor = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );
    __equalize_size( $self, $toor );

    for (my $i = 0; $i < @$self; $i++ ) {
        $$self[ $i ] |= $$toor[ $i ];
    }

    return $self->_pack;
}

=head2 $hi->bit_xor( $value )

C<bit_xor()> does a bitwise or on itself and C<$value>.

    $hi->bit_xor( 127 ); # $hi ^= 127

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub bit_xor {
    my( $self, $value ) = @_;

    my $toxor = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );
    __equalize_size( $self, $toxor );

    for (my $i = 0; $i < @$self; $i++ ) {
        $$self[ $i ] ^= $$toxor[ $i ];
    }

    return $self->_pack;
}

=head2 $hi->add( $value )

C<add()> adds C<$value> to itself.

    $hi->add( 10 ); # $hi += 10;

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub add {
    my( $self, $value ) = @_;

    my $toadd = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );
    __equalize_size( $self, $toadd );

    my $reg = 0;
    tie my @r, 'Hugeint::ByteArray';
    for ( my $i = 0; $i < @$self; $i++ ) {
        my $oreg = $reg;
        $reg += $$self[ $i ] + $$toadd[ $i ];
        push @r, $reg % 256;
        $reg = int( $reg / 256 );
    }

    @$self = @r;
    return $self->_pack;
}

=head2 $hi->subtract( $value )

C<subtract()> subtracts C<$value> from itself.

    $hi->subtract( 10 ); # $hi -= 10;

This is impemented as C<< $hi += -$value >>

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub subtract {
    my( $self, $value ) = @_;
    my $toadd = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );

    $toadd->negate;
    $self->add( $toadd );

    return $self;
}

=head2 $hi->multiply( $value )

C<multiply()> multiplies itself with C<$value>.

    $hi->multiply( 10 ); # $hi *= 10;

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub multiply {
    my( $self, $value ) = @_;

    my $mulby = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );
    __equalize_size( $self, $mulby );

    ( my $s_is_neg = $self->is_negative  ) and $self->negate;
    ( my $m_is_neg = $mulby->is_negative ) and $mulby->negate;

    tie my @R, 'Hugeint::ByteArray';

    for ( my $i = 0; $i < @$self; $i++ ) {
        my $h = 0;
        for ( my $j = 0; $j < @$mulby; $j++ ) {
            $h += $R[ $i + $j ] + $self->[ $i ] * $mulby->[ $j ];
            $R[ $i + $j ] = $h % 256;
            $h = int $h / 256;
        }

        while ( $h > 0 ) {
            push @R, $h % 256;
            $h = int $h / 256;
        }
    }

    ( $R[-1] & 128 ) and push @R, 0; # defensive, but make sure it is positive!
    @$self = @R;
    ( $s_is_neg xor $m_is_neg ) and $self->negate;
    $self->_pack;
}

=head2 $hi->divide( $value )

C<divide()> divides itself by C<$value>. The method differs from the
others as it returns the B<rest> after dividing and not the quotient.

    my $r = $hi->divide( 10 ); # my $r = $hi % 10; $hi /= 10;

B<NOTE>: C<$value> will be converted to I<Hugeint> internally if needed.

=cut

sub divide {
    my( $self, $value ) = @_;

    my $divby = ref $value eq __PACKAGE__
        ? $value->copy
        : __PACKAGE__->new( $value );

    $divby->is_zero and do {
        require Carp;
        Carp::croak( "Illegal division by zero" );
    };

    my $Q = $self->copy;
    my $D = $divby->copy;
    ( my $Q_is_neg = $Q->is_negative ) and $Q->negate;
    ( my $D_is_neg = $D->is_negative ) and $D->negate;
    __equalize_size( $Q, $D );

    if ( hi_compare( $Q, $D ) < 0 ) { # $self < $value
        $self->assign( 0 );
        return $Q;
    }

    my $R = __PACKAGE__->new( 0 );
    while ( hi_compare( $Q, $D ) >= 0 ) {

        # Shift the divisor to the left until it fits
        my $wd = $D->copy;
        while ( hi_compare( $Q, $wd ) >= 0 ) { $wd->shift_left( 1 ) }
        $wd->shift_right( 1 );

        # Find the partial quotient with binsearch
        my($min, $max ) = ( 0, 256 );
        my( $td, $try, $cmp );
        while (  $min < $max ) {
            $try = int( ($max+$min)/2 );
            ( $td = $wd->copy )->multiply( $try );
            $cmp = hi_compare( $Q, $td );

            $cmp == 0 and last;
            $cmp <  0 and $max = $try;
            $cmp >  0 and $min = $try + 1;
            $min >= $max and $cmp < 0 and $try--;
        }
        ( $td = $wd->copy )->multiply( $try );

        $Q->subtract( $td );
        $R->add( $try );

        hi_compare( $wd, $D ) > 0 and $R->shift_left( 1 );
        $Q->_pack;
    }
    ( $Q_is_neg xor $D_is_neg ) and $Q->add( 1 ), $R->negate;
    $D_is_neg and $Q->negate;
    @$self = @$R;
    $self->_pack;
    return $Q->_pack;
}

=head2 $hi->power( $exponent )

C<power()> raises itself by the power of the non-negative integer
C<$exponent>.

    $hi->power( 10 ); # $hi = $hi ** 10;

=cut

sub power {
    my( $self, $exponent ) = @_;
    defined $exponent or $exponent = 1;

    $exponent == 0 and return $self->assign( 1 );
    ($exponent < 1 or int( $exponent ) != $exponent)  and do {
        require Carp;
        Carp::croak( "$exponent should be a non-negative integer!" );
    };

    my $base = $self->copy;

    # reduce to a series of simple operations: 
    # multiply by base; or square
    my @ops = ( );
    while ( $exponent > 1 ) {
        if ( $exponent % 2 == 0 ) { # square $hi and div $exp by 2
            push @ops, 's';
            $exponent = int( $exponent / 2 );
        } else {
            push @ops, 'b';
            $exponent--;
	}
    }

    # handle the operations in reverse order
    for my $op ( reverse @ops ) {
        if ( $op eq 's' ) { # square $hi
            $self->multiply( $self );
        } else { # $hi *= $base
            $self->multiply( $base );
        }
    }

    return $self->_pack;
}

=head1 OVERLOADING OPERATIONS

=begin private

=head2 overload_swap_copy( $a1, $a2[, $a3] )

C<overload_swap_copy()> handles the argumenthandling for overload.

=end private

=cut

sub overload_swap_copy {
    my( $a1, $a2, $a3 ) = @_;
    my( $x, $y );
    if ( $a3 ) {
        $x = __PACKAGE__->new( $a2 );
        $y = __PACKAGE__->new( $a1 );
    } else {
        $x = __PACKAGE__->new( $a1 );
        $y = __PACKAGE__->new( $a2 );
    }
    ( $x, $y );
}

=head2 hi_bit_and( $a1, $a2 [, $a3] )

C<hi_bit_and()> is supplied for overloading the '&' operator.

    my $hi = Hugeint->new( 7 );
    my $r1 = $hi & 255;          # $r1 isa(Hugeint) = 7

=cut

sub hi_bit_and {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->bit_and( $y );
}

=head2 hi_bit_or( $a1, $a2 [, $a3] )

C<hi_bit_or()> is supplied for overloading the '|' operator.

    my $hi = Hugeint->new( 7 );
    my $r1 = $hi | 248;          # $r1 isa(Hugeint) = 255

=cut

sub hi_bit_or {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->bit_or( $y );
}

=head2 hi_bit_xor( $a1, $a2 [, $a3] )

C<hi_bit_xor()> is supplied for overloading the '^' operator.

    my $hi = Hugeint->new( 7 );
    my $r1 = $hi ^ 255;          # $r1 isa(Hugeint) = 248

=cut

sub hi_bit_xor {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->bit_xor( $y );
}

=head2 hi_add( $a1, $a2, $a3 )

C<hi_add()> is supplied for overloading the '+' operator.

    my $h1 = Hugeint->new( 10 );
    my $r1 = $h1 + 20;           # $r1 isa(Hugeint)

=cut

sub hi_add($$;$) {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->add( $y );
}

=head2 hi_subtract( $a1, $a2, $a3 )

C<hi_subtract()> is supplied for overloading the '-' operator.

    my $h1 = Hugeint->new( 20 );
    my $r1 = $h1 - 10;           # $r1 isa(Hugeint)

=cut

sub hi_subtract($$;$) {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->subtract( $y );
}

=head2 hi_multiply( $a1, $a2, $a3 )

C<hi_multiply()> is supplied for overloading the '*' operator.

=cut

sub hi_multiply($$;$) {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->multiply( $y );
}

=head2 hi_divide( $a1, $a2, $a3 )

C<hi_divide()> is supplied for overloading the '/' operator.

=cut

sub hi_divide($$;$) {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->divide( $y );
    $x
}

=head2 hi_modulo( $a1, $a2, $a3 )

C<hi_modulo()> is supplied for overloading the '%' operator.

=cut

sub hi_modulo($$;$) {
    my( $x, $y ) = overload_swap_copy( @_ );

    $x->divide( $y );
}

=head2 hi_power( $a1, $a2 )

C<hi_power()> is supplied for overloading the '**' operator.

=cut

sub hi_power {
    my( $x, $y ) = ( $_[0]->copy, $_[1] );

    $x->power( $y );
}

=head2 hi_compare( $a1, $a2[, $a3] )

C<hi_compare()> implements the spaceship C<< <=> >> opreator. It is
slightly more optimized for cases that do not change the value.

=cut

sub hi_compare {
    my( $a1, $a2, $a3 ) = @_;
    # see overload documentation: always ref $a1 eq __PACKAGE__!!!
    my( $x, $y );
    if ( $a3 ) { # args swapped when $a2 is not a ref
        $x = __PACKAGE__->new( $a2 );
        $y = $a1->copy;
    } else {
        $x = $a1->copy;
        $y = ref $a2 eq __PACKAGE__ ? $a2->copy : __PACKAGE__->new( $a2 );
    }

    my $x_is_neg = $x->is_negative;
    my $y_is_neg = $y->is_negative;

    if ( $x_is_neg xor $y_is_neg ) { # one negative, one positive
        $x_is_neg and return -1;
        return 1;
    }

#    $x_is_neg and $x->negate;
#    $y_is_neg and $y->negate;

    __equalize_size( $x, $y );

    my $csx = $x->_charstr;
    my $csy = $y->_charstr;
    return $csx cmp $csy;

#    my $i = $#{ $x };
#    while ( $i > 0 && $x->[ $i ] == $y->[ $i ] ) { $i-- }
#
#    if ( $x_is_neg ) { # both negative
#        $x->[ $i ] > $y->[ $i ] and return -1;
#        $x->[ $i ] < $y->[ $i ] and return 1;
#        return 0;
#    } else { # both positive
#        $x->[ $i ] > $y->[ $i ] and return 1;
#        $x->[ $i ] < $y->[ $i ] and return -1;
#        return 0;
#    }
}

=head1 COMPARISON OPERATIONS

=head2 $hi->is_negative

Returns a boolean that indicats if the instances value is negative.

=cut

sub is_negative { $_[0]->[-1] & 128 }


=head2 $hi->is_zero

Returns a boolean that indicats if the instances value is zero.

=cut

sub is_zero {
    my $self = shift;
    my $value = $self->_charstr;
    my $zero  = "\00" x length $value;

    ($value eq $zero );
}

=head2 $hi->compare( $value )
    
C<compare()> actually calls the more general C<hi_compare()>.

=cut

sub compare { hi_compare( @_ ) }

=head1 CONVERSION BETWEEN BINARY AND STRING

=head2 $hi->assign( [$str_value] )

Initialize the instance with C<$str_value>. If the initial value is
omitted, it will be initialized with zero (0).

=cut

sub assign {
    my( $self, $value ) = @_;
    defined $value or $value = '0';

    if ( ref $value eq __PACKAGE__ ) {
        @$self = @$value;
        return $self->_pack;
    }

    $value =~ tr/ //d; # remove all spaces
    $value =~ tr/_//d; # remove all '_'

    my( $ops, $digits, $trailing ) =
        $value =~ m/^([+-]*)([0123456789]+)(.*)$/;
    defined $ops or $ops = "";

    $ops =~ tr/-//cd;
    my $is_negative = $ops =~ m/^-(?:--)*$/; # odd number of '-'

    ( $trailing || !defined $digits ) &&
    ( $^W || warnings::enabled( 'numeric' ) ) and
        warnings::warn( "'$_[1]' isn't numeric" );

    defined $digits or $digits = '0'; # at least initialize to 0

    # Convert the 10-base string to a 256-base little-endian byte array
    @$self = ( );
    while ( $digits ne '' ) {
        push @$self, __div_mod_256( $digits );
    }

    # Make sure we have a positive number
    ( $self->[ -1 ] & 128 ) and push @$self, 0; 

    $is_negative and $self->negate;
    return $self->_pack;
}

=head2 $hi->as_string( )

C<as_string()> converts the binary representation of the instance to a
decimal string.
    
=cut
    
sub as_string {
    my $self = $_[0]->copy;
    my $is_neg = $self->is_negative;
    $is_neg and $self->negate;

    my( $r, $s ) = ( 0, '' );
    until ( $self->is_zero ) {
        $r = __div_mod_100( $self );
        $s = __str_100( $r ) . $s;
    }
    length $s == 0 and $s = '0';
    $s =~ s/^0*(?!$)//;
    $is_neg and substr $s, 0, 0, '-';
    return $s
}

=head2 $hi->as_binary( )

C<as_binary()> returns a string with a 2-complement binary
represention of the value of this Hugeint instance. The representation
will be B<big-endian> (as we read decimal numbers) and will have a
space between each byte.

    print $hi->assign( 1 )->as_binary;  # '00000000 00000001'

    print $hi->assign( -1 )->as_binary; # '11111111 11111111'

=cut

sub as_binary {
    my $self = shift;

    return join " ", map sprintf( "%08b", $_ ) => reverse @$self;
}

=head2 __div_mod_256( $value )

C<__dic_mod_256()> B<changes> C<$value> to C<< $value /= 256 >> and
returns C<$value % 256>.

C<$value> is a string representing a decimal number.

B<NOTE>: This changes the value of the calling value as a side effect!

=end private

=cut

sub __div_mod_256 {
    my( $value ) = @_;
    my( $q, $r ) = ( '', 0 );
    while ( $value ne '' ) {
        $r = 10 * $r + ord( substr $value, 0, 1 ) - ord( '0' );
        $value =~ s/^.//;
        $q .= chr( int( $r / 256 ) + ord( '0' ) );
        $r %= 256;
    }
    $q =~ s/^0+//;
    $_[0] = $q;
    return $r;
}

=begin private

=head2 __div_mod_100( $hugeint )

C<__div_mod_100()> B<changes> C<$hugeint> to C<< $hugeint /= 100 >> and
returns C<$hugeint % 100>.

C<$hugeint> is a I<ByteArry> that represents a binary number.

=end private

=cut

sub __div_mod_100 {
    my( $hugeint ) = @_;
    my $r = 0;
    tie my @Q, 'Hugeint::ByteArray';
    for my $byte ( reverse @$hugeint ) {
        $r = 256 * $r + $byte;
        unshift @Q, int( $r / 100 );
        $r %= 100;
    }
    @{ $_[0] } = @Q;
    return $r;
}

=begin private

=head2 __str_100( $value )

C<__str_100()> returns a string representing a number between 0 and 99.

=end private

=cut

sub __str_100 {
    my( $value ) = @_;

    return chr( int( $value / 10 ) + ord( '0' ) ) .
           chr( ($value % 10) + ord( '0' ) );
}

=begin private

=head2 _charstr

This method returns a char string representation of the Hugeint.

=end private

=cut

sub _charstr {
    my $self = shift;

    return join "", map chr( $_ ) => reverse @$self;
}

1;
