#! perl
use warnings;
use strict;

my @assigns = qw/
    0 1 255 256 65535 65536
    -1 -65535 -12345678901234567890 12345678901234567890
/;
use Test::More;
use Test::Warn;

plan tests =>  1                 # use_ok
            +  2 * @assigns + 25 # assignment
            + 11                 # addition
            +  6                 # subtraction
            +  8                 # multiplication
            + 11                 # division
            + 10                 # power
            +  9                 # bitwise stuff
            +  3                 # compare stuff
            +  6                 # misc stuff
;

# Begin testing the object interface

use_ok 'Hugeint';

{ # assignment, conversion
    for my $val ( @assigns ) {
        my $hi = Hugeint->new( $val );
        isa_ok $hi, 'Hugeint', "($val)";
        is $hi->as_string, $val, "conversion ok";
    }
    my $pval = 12345;
    for my $cnt ( 0..5 ) {
        my $minus = '-' x $cnt;
        my $isneg = $cnt % 2 == 1;
        my $hi = Hugeint->new( "$minus$pval" );
        isa_ok $hi, 'Hugeint', "($minus$pval)";
        my $val  = $hi->as_string;
        my $cval = $isneg ? -$pval : $pval; 
        is $val, $cval, "($minus$pval) == $cval";
    }

    {
        my $hi = Hugeint->new;
        isa_ok $hi, 'Hugeint';
        $hi->assign;
        ok $hi->is_zero, "still 0 after undef assignment";
    }

    { # These should warn about invalid value
        my $hi;
        warning_like {
            $hi = Hugeint->new( '1garbage' );
            isa_ok $hi, 'Hugeint';
            is $hi->as_string, '1', "assign with invalid value leading digits";
        } qr/'1garbage' isn't numeric/, "warnings 'numeric' (value)";

        warning_like {
            $hi->assign( 'not a number' );
            is $hi->as_string, '0', "assign with invalid value";
        } qr/'not a number' isn't numeric/, "warnings 'numeric' (no value)";

        {
            no warnings 'numeric';
            $hi->assign( '1garbage' );
            is $hi->as_string, '1', "assign with invalid value leading digits";
    
            $hi->assign( 'not a number' );
            is $hi->as_string, '0', "assign with invalid value";

            local $^W = 1;
            warning_like {
                $hi->assign( '1garbage' );
                is $hi->as_string, '1', "assign invalid value leading digits";
            } qr/'1garbage' isn't numeric/, "warnings 'numeric' (value)";
    
            warning_like {
                $hi->assign( 'not a number' );
                is $hi->as_string, '0', "assign with invalid value";
            } qr/'not a number' isn't numeric/, "warnings 'numeric' (no value)";
        }
    }
}

{ # addition
    my $hi = Hugeint->new;
    isa_ok $hi, 'Hugeint';
    my $val = $hi->as_string;
    is $val, '0', "new instance initialized to 0";

    
    $hi->add( 255 ); $val = $hi->as_string;
    is $val, '255', "add(255): " . $val;

    $hi->add( 1 ); $val = $hi->as_string;
    is $val, '256', "add(1) " . $val;

    $hi->assign( 0 ); $val = $hi->as_string;
    is $val, '0', "assign(0) " . $val;
    
    $hi->add( 65535 ); $val = $hi->as_string;
    is $val, '65535', "add(65535) " . $val;

    $hi->add( 1 ); $val = $hi->as_string;
    is $val, '65536', "add(1) " . $val;

    $hi->add(-1); $val = $hi->as_string;
    is $val, '65535', "add(-1) " . $val;

    $hi->assign( -10 ); $val = $hi->as_string;
    is $val, '-10', "assign(-10) " . $val;

    $hi->add( -10 ); $val = $hi->as_string;
    is $val, '-20', "add(-10) $val";

    $hi->add( 30 ); $val = $hi->as_string;
    is $val, '10', "add(30) $val";
}

{ # subtraction
    my $hi = Hugeint->new( 100 );
    isa_ok $hi, 'Hugeint';
    my $val = $hi->as_string;
    is $val, '100', "assign(100) $val";

    $hi->subtract( 50 ); $val = $hi->as_string;
    is $val, '50', "subtract(50) $val";

    $hi->subtract( 100 ); $val = $hi->as_string;
    is $val, '-50', "subtract(100) $val";

    $hi->subtract( -75 ); $val = $hi->as_string;
    is $val, '25', "subtract(-75) $val";

    $hi->subtract( Hugeint->new( 25 ) ); $val = $hi->as_string;
    is $val, '0', "subtract( hi(25) ) $val";
}

{ # multiplication
    my $hi = Hugeint->new( 3 );
    isa_ok $hi, 'Hugeint';
    my $val = $hi->as_string;
    is $val, '3', "assign(3) $val";

    $hi->multiply( 3 ); $val = $hi->as_string;
    is $val, '9', "multiply(3) $val";

    $hi->multiply( Hugeint->new( 3 ) ); $val = $hi->as_string;
    is $val, '27', "multiply( hi(3) ) $val";

    $hi->multiply( -3 ); $val = $hi->as_string;
    is $val, '-81', "multiply(-3) $val";

    $hi->multiply( 81 ); $val = $hi->as_string;
    is $val, '-6561', "multiply(-81) $val";

    $hi->multiply(-81); $val = $hi->as_string;
    is $val, '531441', "multiply(-81) $val";

    $hi->assign( 65535 );
    $hi->multiply( 65535 ); $val = $hi->as_string;
    is $val, '4294836225', "multiply(65535) $val";
}

{ # division
    my $hi = Hugeint->new( 531441 );
    isa_ok $hi, 'Hugeint';
    my $val = $hi->as_string;
    is $val, '531441', "assign(531441) $val";

    $hi->divide( 81 ); $val = $hi->as_string;
    is $val, '6561', "divide(81) $val";

    $hi->divide( -81 ); $val = $hi->as_string;
    is $val, '-81', "divide(-81) $val";

    $hi->divide( 3 ); $val = $hi->as_string;
    is $val, '-27', "divide(-3) $val";

    $hi->divide( Hugeint->new( -3 ) ); $val = $hi->as_string;
    is $val, '9', "divide( hi(3) ) $val";

    my $r = $hi->divide( 5 ); $val = $hi->as_string;
    is $r, 4, "9 % 5 $r";
    is $val, '1', "divide(5) $val";

    $hi->divide( 10 ); $val = $hi->as_string;
    ok $hi->is_zero, "divide(10) $val";

    $hi->assign( 100 );
    eval { $hi->divide( 0 ) };
    like $@, qr/Illegal division by zero at/, "Cannot divide by zero; fatal";

    $hi->assign( 513 );
    $hi->divide( 1 ); $val = $hi->as_string;
    is $val, '513', "divide(1) $val";
}

{ # power
    my $hi = Hugeint->new( 3 );
    isa_ok $hi, 'Hugeint';

    my $val = $hi->as_string;
    is $val, '3', "assign(3) $val";

    $hi->power( 9 ); $val = $hi->as_string;
    is $val, '19683', "power(9) $val";

    $hi->assign( 2 );
    $hi->power( 128 ); $val = $hi->as_string;
    is $val, '340282366920938463463374607431768211456', "2^128 = $val";

    $hi->power; $val = $hi->as_string;
    is $val, '340282366920938463463374607431768211456', "2^128 = $val";

    $hi->power( 0 ); $val = $hi->as_string;
    is $val, '1', "power(0) $val";

    $hi->assign( -3 );
    $hi->power( 3 ); $val = $hi->as_string;
    is $val, '-27', "-3^3 = -27";

    $hi->power( 2 ); $val = $hi->as_string;
    is $val, '729', "-27^2 = $val";

    eval { $hi->power( 0.5 ) };
    like $@, qr/should be a non-negative integer!/, "Need an iteger argument";

    eval { $hi->power( 1.5 ) };
    like $@, qr/should be a non-negative integer!/, "Need an iteger argument";
}

{ # bit operators
    my $hi = Hugeint->new( 1 );

    # ~ $hi + 1 => - $hi
    my $an = $hi->copy;
    isa_ok $an, 'Hugeint', "arithmatic";
    my $bn = $hi->copy;
    isa_ok $bn, 'Hugeint', "bitwise";

    $an->negate;               my $av = $an->as_string;
    $bn->bit_negate->add( 1 ); my $bv = $bn->as_string;
    is $av, $bv, "~1 + 1($bv) == - 1($av)";

    $hi->shift_left( 2 ); my $val = $hi->as_string;
    is $val, '65536', "1 << 2 == $val";

    $hi->assign( 65534 );
    is $hi->as_binary, '00000000 00000000 11111111 11111110', "check as_binary";

    $hi->bit_and( 127 ); $val = $hi->as_string;
    is $val, '126', "65534 & 127 = $val"; 

    $hi->bit_or( 127 ); $val = $hi->as_string;
    is $val, '127', "126 | 127 = $val";

    # due to internal representation, this may go wrong
    $hi->assign( 65535 )->bit_and( $bn ); $val = $hi->as_string;
    is $val, '65535', "65535 & -1 = $val";

    $hi->assign( 65535 )->bit_or( $bn ); $val = $hi->as_string;
    is $val, '-1', "65535 | -1 = $val"; 
}

{ # compare stuff
    my $hi = Hugeint->new( 10 );
    isa_ok $hi, 'Hugeint';

    is $hi->compare( 20 ),                 -1, "10 < 20";
    is $hi->compare( Hugeint->new( 20 ) ), -1, "10 < hi(20)";
}

{ # misc to trigger edgecases in the implentation
    my $hi = Hugeint->new;
    isa_ok $hi, 'Hugeint';
    @$hi = ();
    $hi->_pack;
    is $hi->as_string, '0', "Edgecase 1; internal value lost";

    @$hi = ( 1, 0, 255, 255, 255, 255, 255, 255 );
    $hi->_pack;
    is $hi->as_string, '-65535', "Edgecase 2; shrink negative value";

    $hi->assign( 3 )->multiply( Hugeint->new( '21846' ) );
    is $hi->as_string, '65538', "Edgecase 3; trigger register overload in *";

    $hi->assign( 256 )->multiply( 128 );
    is $hi->as_string, '32768', "Edgecase 4; trigger extending in *";

    @$hi = ( 1 );
    $hi->shift_right( 1 );
    is $hi->as_string, '0', "Edgecase 5; trigger short number extendig in >>";
}
